/*
 * NET		An implementation of the SOCKET network access protocol.
 *		This module implements the SLIP protocol for the kernel-
 *		based devices like TTY.  It interfaces between a raw TTY,
 *		and the kernel's NET protocol layers (via DDI).
 *
 *		Van Jacobsen Header Compression functions.
 *
 * Version:	@(#)compress.c	1.7.1	04/25/94
 *
 * Authors:	Van Jacobson, <van@helios.ee.lbl.gov>, Dec 31, 1989
 *		See the `COPYRIGHT' file in this directory for the
 *		legal status and version info of this code.
 *
 *		Katie Stevens, <dkstevens@ucdavis.edu>, 01-31-90
 *		Phil Karn, <karn@ka9q.ampr.org>, 11-02-90
 *		Bill Simpson, <Bill_Simpson@um.cc.umich.edu>, 02-91
 *		Chuck Hedrick, <hedrick@klinzhai.rutgers.edu>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *		Jim Seagrave, <jes@grendel.demon.co.uk>
 *		Steve Bour, <jsbour@ualr.edu>
 */
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/timer.h>
#include <linux/ddi.h>
#include <linux/skbuff.h>
#include <linux/in_systm.h>
#include <linux/in.h>		/* for IPPROTO_TCP, ntohs() et al */
#include <linux/if.h>
#include <linux/if_ppp.h>
#include <linux/ip.h>
#include <linux/tty.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/tty.h>
#include "ppp.h"
#include "global.h"
#include "compress.h"


/* Put a short in host order into a char array in network order. */
static unsigned char *
put16(unsigned char *cp, unsigned short int x)
{
  *cp++ = (x >> 8);
  *cp++ = x;

  return(cp);
}


/* Pull a 16-bit integer in host order from buffer in network byte order. */
static short int
pull16(unsigned char **cpp)
{
  short int rval;

  rval = *(*cpp)++;
  rval <<= 8;
  rval |= *(*cpp)++;
  return(rval);
}


/* Encode a number. */
static unsigned char *
encode(unsigned char *cp, unsigned int n)
{
  if (n >= 256 || n == 0) {
	*cp++ = 0;
	cp = put16(cp, n);
  } else {
	*cp++ = n;
  }
  return(cp);
}


/* Decode a number. */
static long int
decode(unsigned char **cpp)
{
  long int x;

  x = (long int) *(*cpp)++;
  if (x == 0) {
	return(pull16(cpp) & 0xFFFF);	/* pull16 returns -1 on error */
  } else {
	return(x & 0xFF);		/* -1 if PULLCHAR returned error */
  }
}


/* Dump a VJ header. */
#ifdef PPP_DEBUG
void
pphc_print(unsigned char *cp, int len)
{
  unsigned char *ptr;
  short int csum;

  ptr = (cp + 2);
  csum = pull16(&ptr);
  printk("PPP/VJ: type=0x%02X changes=0x%02X sum=0x%X len=%d\n",
		(*cp & 0377), (*(cp + 1) & 0377), (int) csum, len);
}
#endif	/* PPP_DEBUG */


/*
 * This routine does all the checksum computations that don't
 * require anything special (like copying or special headers.)
 */
static unsigned short
pphc_ip_compute_csum(unsigned char *buff, int len)
{
  unsigned long sum = 0;

  /* Do the first multiple of 4 bytes and convert to 16 bits. */
  if (len > 3) {
	__asm__("clc\n"
	        "1:\t"
	    	"lodsl\n\t"
	    	"adcl %%eax, %%ebx\n\t"
	    	"loop 1b\n\t"
	    	"adcl $0, %%ebx\n\t"
	    	"movl %%ebx, %%eax\n\t"
	    	"shrl $16, %%eax\n\t"
	    	"addw %%ax, %%bx\n\t"
	    	"adcw $0, %%bx"
	        : "=b" (sum) , "=S" (buff)
	        : "0" (sum), "c" (len >> 2) ,"1" (buff)
	        : "ax", "cx", "si", "bx" );
  }
  if (len & 2) {
	__asm__("lodsw\n\t"
	    	"addw %%ax, %%bx\n\t"
	    	"adcw $0, %%bx"
	        : "=b" (sum), "=S" (buff)
	        : "0" (sum), "1" (buff)
	        : "bx", "ax", "si");
  }
  if (len & 1) {
	__asm__("lodsb\n\t"
	    	"movb $0, %%ah\n\t"
	    	"addw %%ax, %%bx\n\t"
	    	"adcw $0, %%bx"
	        : "=b" (sum), "=S" (buff)
	        : "0" (sum), "1" (buff)
	        : "bx", "ax", "si");
  }
  sum =~sum;
  return(sum & 0xFFFF);
}


/* Check the header of an incoming IP datagram. */
static int
pphc_ip_csum(struct ip *iph)
{
  if (pphc_ip_compute_csum((unsigned char *)iph,
		      (iph->ip_hl * sizeof(unsigned long))) == 0) return(0);
  return(1);
}


void
pphc_i_status(struct compress *comp)
{
  if (comp != NULL) {
	printk("\t%ld Cmp, %ld Uncmp, %ld Bad, %ld Tossed\n",
		comp->pp->stats.rcomp, comp->pp->stats.runcomp,
		comp->pp->stats.rerrors, comp->pp->stats.tossed);
  }
}


void
pphc_o_status(struct compress *comp)
{
  if (comp != NULL) {
	printk("\t%ld Cmp, %ld Uncmp, %ld AsIs, %ld NotTCP\n",
		comp->pp->stats.scomp, comp->pp->stats.suncomp,
		0 /* FIXME: ??? */, comp->pp->stats.sothers);
#if 0
	printk("\t%10ld Searches, %10ld Misses\n",
		comp->pp->stats.searches, comp->pp->stats.misses);
#endif
  }
}


int
pphc_toss(struct compress *comp)
{
  if (comp == NULL) return(0);

  comp->flags |= SLF_TOSS;
  return(0);
}


/* Free a compression data structure. */
void
pphc_free(struct compress *comp)
{
  if (comp == NULL) return;

  if (comp->rstate != NULL) kfree(comp->rstate);

  if (comp->tstate != NULL) kfree(comp->tstate);

  kfree_s(comp, sizeof(struct compress));
}


/*
 * Initialize compression data structure
 * 'slots' must be in range 0 to 255 (zero meaning no compression)
 */
struct compress *
pphc_init(struct ppp *pp, int rslots, int tslots)
{
  struct compress *comp;
  struct cstate *ts;
  short i;

  comp = (struct compress *) kmalloc(sizeof(struct compress), GFP_KERNEL);
  if (comp == NULL) return(NULL);

  memset(comp, 0, sizeof(struct compress));
  if (rslots > 0 && rslots < 256) {
	comp->rstate = (struct cstate *)
			kmalloc(rslots * sizeof(struct cstate), GFP_KERNEL);
	if (comp->rstate == NULL) return(NULL);
	memset(comp->rstate, 0, rslots * sizeof(struct cstate));
	comp->rslot_limit = rslots - 1;
  }

  if (tslots > 0 && tslots < 256) {
	comp->tstate = (struct cstate *)
			kmalloc(tslots * sizeof(struct cstate), GFP_KERNEL);
	if (comp->tstate == NULL) return(NULL);
	memset(comp->tstate, 0, rslots * sizeof(struct cstate));
	comp->tslot_limit = tslots - 1;
  }
  comp->xmit_oldest = 0;
  comp->xmit_current = 255;
  comp->recv_current = 255;
  comp->pp = pp;

  /*
   * don't accept any packets with implicit index until we get
   * one with an explicit index.  Otherwise the uncompress code
   * will try to use connection 255, which is almost certainly
   * out of range.
   */
  comp->flags |= SLF_TOSS;

  if (tslots > 0) {
	ts = comp->tstate;
	for(i = comp->tslot_limit; i > 0; --i) {
		ts[i].connid = i;
		ts[i].next = &(ts[i - 1]);
	}
	ts[0].next = &(ts[comp->tslot_limit]);
	ts[0].connid = 0;
  }
  return(comp);
}


/* 
 * icp and isize are the original packet.
 * ocp is a place to put a copy if necessary.
 * cpp is initially a pointer to icp.  If the copy is used,
 *    change it to ocp.
 */
int
pphc_compress(struct compress *comp, unsigned char *icp, int isize,
	      unsigned char *ocp, unsigned char **cpp, int compress_cid)

{
  unsigned char new_seq[16];
  struct cstate *ocs = &(comp->tstate[comp->xmit_oldest]);
  struct cstate *lcs = ocs;
  struct cstate *cs = lcs->next;
  unsigned char *cp = new_seq;
  struct tcphdr *th, *oth;
  struct ip *ip;
  unsigned long deltaS, deltaA;
  short changes = 0;
  int hlen;

  ip = (struct ip *) icp;

  /* Bail if this packet isn't TCP, or is an IP fragment */
  if (ip->ip_p != IPPROTO_TCP ||
      (ntohs(ip->ip_off) & IP_OFFSET) || (ip->ip_off & 32)) {
	DPRINTF(("comp: noncomp 1 %d %d %d\n", ip->ip_p, 
					ntohs(ip->ip_off), ip->ip_off));
#if 0
	/* Send as regular IP */
	if (ip->ip_p != IPPROTO_TCP) comp->pp->stats.sothers++;
	  else comp->pp->stats.tcp++;
#endif
	return(isize);
  }

  /* Extract TCP header */
  th = (struct tcphdr *)
	(((unsigned char *)ip) + (ip->ip_hl * sizeof(unsigned long)));
  hlen = (ip->ip_hl * sizeof(unsigned long)) +
	 (th->doff * sizeof(unsigned long));

  /*
   * Bail if the TCP packet isn't `compressible' (i.e.,
   * ACK isn't set or some other control bit is set).
   */
  if (th->syn || th->fin || th->rst || !(th->ack)) {
	DPRINTF(("comp: noncomp 2 %x %x %d %d %d %d\n", ip, th, 
					th->syn, th->fin, th->rst, th->ack));
	/* TCP connection stuff; send as regular IP */
#if 0
	comp->pp->stats.tcp++;
#endif
	return(isize);
  }

  /*
   * Packet is compressible -- we're going to send either a COMPRESSED_TCP
   * or UNCOMPRESSED_TCP packet.  Either way, we need to locate (or create)
   * the connection state.
   *
   * States are kept in a circularly linked list with xmit_oldest pointing
   * to the end of the list.  The list is kept in lru order by moving a
   * state to the head of the list whenever it is referenced.  Since the
   * list is short and, empirically, the connection we want is almost al-
   * ways near the front, we locate states via linear search.  If we don't
   * find a state for the datagram, the oldest state is (re-)used.
   */
  for ( ; ; ) {
	if (ip->ip_src.s_addr == cs->cs_ip.ip_src.s_addr &&
	    ip->ip_dst.s_addr == cs->cs_ip.ip_dst.s_addr &&
	    th->source == cs->cs_tcp.source &&
	    th->dest == cs->cs_tcp.dest) goto found;

	/* If current equal oldest, at end of list. */
	if (cs == ocs) break;
	lcs = cs;
	cs = cs->next;
#if 0
	comp->pp->stats.searches++;
#endif
  };

  /*
   * Didn't find it -- re-use oldest cstate.  Send an
   * uncompressed packet that tells the other side what
   * connection number we're using for this conversation.
   *
   * Note that since the state list is circular, the oldest
   * state points to the newest and we only need to set
   * xmit_oldest to update the lru linkage.
   */
#if 0
  comp->pp->stats.misses++;
#endif
  comp->xmit_oldest = lcs->connid;
  DPRINTF(("comp: not found\n"));
  goto uncompressed;

found:
  /* Found it -- move to the front on the connection list. */
  if (lcs == ocs) {
	/* found at most recently used */
  } else if (cs == ocs) {
	/* found at least recently used */
	comp->xmit_oldest = lcs->connid;
  } else {
	/* more than 2 elements */
	lcs->next = cs->next;
	cs->next = ocs->next;
	ocs->next = cs;
  }

  /*
   * Make sure that only what we expect to change changed.
   * Check the following:
   * IP protocol version, header length & type of service.
   * The "Don't fragment" bit.
   * The time-to-live field.
   * The TCP header length.
   * IP options, if any.
   * TCP options, if any.
   * If any of these things are different between the previous &
   * current datagram, we send the current datagram `uncompressed'.
   */
  oth = &cs->cs_tcp;

  if (
      ip->ip_v != cs->cs_ip.ip_v ||
      ip->ip_hl != cs->cs_ip.ip_hl ||
      ip->ip_tos != cs->cs_ip.ip_tos ||
      (ip->ip_off & 64) != (cs->cs_ip.ip_off & 64) ||
      ip->ip_ttl != cs->cs_ip.ip_ttl ||
      th->doff != cs->cs_tcp.doff ||
      (ip->ip_hl > 5 &&
       memcmp(ip + 1, cs->cs_ipopt, ((ip->ip_hl) - 5) * sizeof(unsigned long))
									!= 0) ||
      (th->doff > 5 && memcmp(th + 1, cs->cs_tcpopt, ((th->doff) - 5) *
						sizeof(unsigned long) != 0))
    ) {
	DPRINTF(("comp: incompat\n"));
	goto uncompressed;
  }
	
  /*
   * Figure out which of the changing fields changed.  The
   * receiver expects changes in the order: urgent, window,
   * ack, seq (the order minimizes the number of temporaries
   * needed in this section of code).
   */
  if (th->urg) {
	deltaS = ntohs(th->urg_ptr);
	cp = encode(cp, deltaS);
	changes |= NEW_U;
  } else if (th->urg_ptr != oth->urg_ptr) {
	/*
	 * Argh! URG not set but urp changed -- a sensible
	 * implementation should never do this but RFC793
	 * doesn't prohibit the change so we have to deal
	 * with it.
	 */
	DPRINTF(("comp: urg incompat\n"));
	goto uncompressed;
  }
  if ((deltaS = ntohs(th->window) - ntohs(oth->window)) != 0) {
	cp = encode(cp, deltaS);
	changes |= NEW_W;
  }
  if ((deltaA = ntohl(th->ack_seq) - ntohl(oth->ack_seq)) != 0L) {
	if (deltaA > 0x0000FFFF) goto uncompressed;
	cp = encode(cp, deltaA);
	changes |= NEW_A;
  }
  if ((deltaS = ntohl(th->seq) - ntohl(oth->seq)) != 0L) {
	if (deltaS > 0x0000FFFF) goto uncompressed;
	cp = encode(cp, deltaS);
	changes |= NEW_S;
  }
	
  switch(changes) {
	case 0:
		/*
		 * Nothing changed. If this packet contains data and the
		 * last one didn't, this is probably a data packet following
		 * an ack (normal on an interactive connection) and we send
		 * it compressed.  Otherwise it's probably a retransmit,
		 * retransmitted ack or window probe.  Send it uncompressed
		 * in case the other side missed the compressed version.
		 */
		if (ip->ip_len != cs->cs_ip.ip_len &&
		    ntohs(cs->cs_ip.ip_len) == hlen) break;
		DPRINTF(("comp: retrans\n"));
		goto uncompressed;
		break;

	case SPECIAL_I:
	case SPECIAL_D:
		/*
		 * Actual changes match one of our special case
		 * encodings -- send packet uncompressed.
 		 */
		DPRINTF(("comp: special\n"));
		goto uncompressed;

	case NEW_S|NEW_A:
		if (deltaS == deltaA &&
		    deltaS == ntohs(cs->cs_ip.ip_len) - hlen) {
			/* Special case for echoed terminal traffic. */
			changes = SPECIAL_I;
			cp = new_seq;
		}
		break;

	case NEW_S:
		if (deltaS == ntohs(cs->cs_ip.ip_len) - hlen) {
			/* Special case for data xfer. */
			changes = SPECIAL_D;
			cp = new_seq;
		}
		break;
  }

  deltaS = ntohs(ip->ip_id) - ntohs(cs->cs_ip.ip_id);
  if (deltaS != 1) {
	cp = encode(cp, deltaS);
	changes |= NEW_I;
  }
  if (th->psh) changes |= TCP_PUSH_BIT;

  /*
   * Grab the cksum before we overwrite it below.  Then
   * update our state with this packet's header.
   */
  deltaA = ntohs(th->check);
  memcpy(&cs->cs_ip, ip, sizeof(struct ip));		/* FIXME: 20 */
  memcpy(&cs->cs_tcp, th, sizeof(struct tcphdr));	/* FIXME: 20 */

  /*
   * We want to use the original packet as our compressed packet.
   *
   * (cp - new_seq) is the number of bytes we need for compressed
   * sequence numbers.  In addition we need one byte for the change
   * mask, one for the connection id and two for the tcp checksum.
   * So, (cp - new_seq) + 4 bytes of header are needed.
   */
  deltaS = cp - new_seq;
  if (compress_cid == 0 || comp->xmit_current != cs->connid) {
	cp = ocp;
	*cpp = ocp;
	*cp++ = changes | NEW_C;
	*cp++ = cs->connid;
	comp->xmit_current = cs->connid;
  } else {
	cp = ocp;
	*cpp = ocp;
	*cp++ = changes;
  }
  cp = put16(cp, (short)deltaA);	/* Write TCP checksum */

  /* deltaS is now the size of the change section of the compressed header. */
  DPRINTF(("comp: %x %x %x %d %d\n", icp, cp, new_seq, hlen, deltaS));
  memcpy(cp, new_seq, deltaS);	/* Write list of deltas */
  memcpy(cp + deltaS, icp + hlen, isize - hlen);
  comp->pp->stats.scomp++;
  ocp[0] |= PPP_TYPE_COMP_TCP;
  return(isize - hlen + deltaS + (cp - ocp));

uncompressed:
  /*
   * Update connection state cs & send uncompressed packet (i.e.,
   * a regular ip/tcp packet but with the 'conversation id' we hope
   * to use on future compressed packets in the protocol field).
   */
  memcpy(&cs->cs_ip, ip, sizeof(struct ip));		/* FIXME: 20 */
  memcpy(&cs->cs_tcp, th, sizeof(struct tcphdr));	/* FIXME: 20 */
  if (ip->ip_hl > 5)
	memcpy(cs->cs_ipopt, ip+1, ((ip->ip_hl) - 5) * sizeof(unsigned long));
  if (th->doff > 5)
	memcpy(cs->cs_tcpopt, th+1, ((th->doff) - 5) * sizeof(unsigned long));
  comp->xmit_current = cs->connid;
  comp->pp->stats.suncomp++;
  memcpy(ocp, icp, isize);
  *cpp = ocp;
  ocp[9] = cs->connid;
  ocp[0] |= PPP_TYPE_UNCOMP_TCP;
  return(isize);
}


int
pphc_uncompress(struct compress *comp, unsigned char *icp, int isize)
{
  struct tcphdr *thp;
  struct ip *ip;
  struct cstate *cs;
  unsigned char *cp = icp;
  int changes;
  int len, hdrlen;
  long x;

  /* We've got a compressed packet; read the change byte. */
  comp->pp->stats.rcomp++;
  if (isize < 3) {
	comp->pp->stats.rerrors++;
	DPRINTF(("uncomp: runt\n"));
	return(0);
  }
  changes = *cp++;
  if (changes & NEW_C) {
	/*
	 * Make sure the state index is in range, then grab the state.
	 * If we have a good state index, clear the 'discard' flag.
	 */
	x = *cp++;	/* Read conn index */
	if (x < 0 || x > comp->rslot_limit) goto bad;

	comp->flags &=~ SLF_TOSS;
	comp->recv_current = x;
  } else {
	/*
	 * This packet has an implicit state index.  If
	 * we've had a line error since the last time we
	 * got an explicit state index, we have to toss
	 * the packet.
	 */
	if (comp->flags & SLF_TOSS) {
		comp->pp->stats.tossed++;
		DPRINTF(("uncomp: toss\n"));
		return(0);
	}
  }
  cs = &comp->rstate[comp->recv_current];
  thp = &cs->cs_tcp;
  ip = &cs->cs_ip;

  if ((x = pull16(&cp)) == -1) {	/* Read the TCP checksum */
	DPRINTF(("uncomp: bad tcp chk\n"));
	goto bad;
  }
  thp->check = htons(x);
  thp->psh = (changes & TCP_PUSH_BIT) ? 1 : 0;

  /* 
   * we can use the same number for the length of the saved header and
   * the current one, because the packet wouldn't have been sent
   * as compressed unless the options were the same as the previous one
   */
  hdrlen = (ip->ip_hl * sizeof(unsigned long)) +
	   (thp->doff * sizeof(unsigned long));
  switch(changes & SPECIALS_MASK) {
	case SPECIAL_I:		/* Echoed terminal traffic */
		{
			short i;

			i = ntohs(ip->ip_len) - hdrlen;
			thp->ack_seq = htonl(ntohl(thp->ack_seq) + i);
			thp->seq = htonl(ntohl(thp->seq) + i);
		}
		break;

	case SPECIAL_D:		/* Unidirectional data */
		thp->seq = htonl(ntohl(thp->seq) +
				 ntohs(ip->ip_len) - hdrlen);
		break;

	default:
		if (changes & NEW_U) {
			thp->urg = 1;
			if ((x = decode(&cp)) == -1) {
				DPRINTF(("uncomp: bad U\n"));
				goto bad;
			}
			thp->urg_ptr = htons(x);
		} else {
			thp->urg = 0;
		}
		if (changes & NEW_W) {
			if ((x = decode(&cp)) == -1) {
				DPRINTF(("uncomp: bad W\n"));
				goto bad;
			}	
			thp->window = htons(ntohs(thp->window) + x);
		}
		if (changes & NEW_A) {
			if ((x = decode(&cp)) == -1) {
				DPRINTF(("uncomp: bad A\n"));
				goto bad;
			}
			thp->ack_seq = htonl(ntohl(thp->ack_seq) + x);
		}
		if (changes & NEW_S) {
			if ((x = decode(&cp)) == -1) {
				DPRINTF(("uncomp: bad S\n"));
				goto bad;
			}
			thp->seq = htonl(ntohl(thp->seq) + x);
		}
		break;
  }

  if (changes & NEW_I) {
	if ((x = decode(&cp)) == -1) {
		DPRINTF(("uncomp: bad I\n"));
		goto bad;
	}
	ip->ip_id = htons(ntohs(ip->ip_id) + x);
  } else {
	ip->ip_id = htons(ntohs(ip->ip_id) + 1);
  }

  /*
   * At this point, cp points to the first byte of data in the
   * packet.  Put the reconstructed TCP and IP headers back on the
   * packet.  Recalculate IP checksum (but not TCP checksum).
   */
  len = isize - (cp - icp);
  if (len < 0) goto bad;
  len += hdrlen;
  ip->ip_len = htons(len);
  ip->ip_sum = 0;
  DPRINTF(("uncomp: %d %d %d %d\n", cp - icp, hdrlen, isize, len));

  /*
   * Note from ChuckH to FredK: "this routine'd better behave well
   * when confronted with overlapping copies!" -FvK
   */
  memmove(icp + hdrlen, cp, len - hdrlen);

  /* Copy in the IP header (and options). */
  cp = icp;
  memcpy(cp, ip, sizeof(struct ip));		/* FIXME: 20 */
  cp += sizeof(struct ip);			/* FIXME: 20 */
  if (ip->ip_hl > 5) {
	memcpy(cp, cs->cs_ipopt, ((ip->ip_hl) - 5) * sizeof(unsigned long));
	cp += ((ip->ip_hl) - 5) * sizeof(unsigned long);
  }

#if 1	/* FIXME: Check This Soon */
  ip = (struct ip *) icp;
  ip->ip_sum = pphc_ip_compute_csum((unsigned char *)ip,
		      (ip->ip_hl * sizeof(unsigned long)));
#else
  ((struct ip *)icp)->ip_sum = pphc_ip_csum((struct ip *)icp);
#endif

  /* Copy in the TCP header (and options). */
  memcpy(cp, thp, sizeof(struct tcphdr));	/* FIXME: 20 */
  cp += sizeof(struct tcphdr);
  if (thp->doff > 5) {
	memcpy(cp, cs->cs_tcpopt, ((thp->doff) - 5) * sizeof(unsigned long));
	cp += ((thp->doff) - 5) * sizeof(unsigned long);
  }

  DPRINTF(("\runcomp: change %x len %d\n", changes, len));
  return(len);

bad:
  comp->pp->stats.rerrors++;
  return(pphc_toss(comp));
}


int
pphc_remember(struct compress *comp, unsigned char *icp, int isize)
{
  struct ip *ip;
  struct tcphdr *thp;
  struct cstate *cs;
  unsigned char index;
  short ip_len;

  if (isize < sizeof(struct ip)) {
	/* The packet is shorter than a legal IP header. */
	comp->pp->stats.runts++;
	return(pphc_toss(comp));
  }

  /* Clear the IP header "version" bit field. */
  icp[0] &= 0x4F;

  /* Sneak a peek at the IP header's IHL field to find its length. */
  ip_len = (icp[0] & 0x0F) << 2;
  if (ip_len < sizeof(struct ip)) {
	/* The IP header length field is too small. */
	comp->pp->stats.runts++;
	return(pphc_toss(comp));
  }
  index = icp[9];
  icp[9] = IPPROTO_TCP;
  ip = (struct ip *) icp;

  if (pphc_ip_csum(ip)) {
	/* Bad IP header checksum; discard. */
	comp->pp->stats.rerrors++;
	return(pphc_toss(comp));
  }

  thp = (struct tcphdr *)
	(((unsigned char *)ip) + (ip->ip_hl * sizeof(unsigned long)));
  if (index > comp->rslot_limit) {
	comp->pp->stats.rerrors++;
	return(pphc_toss(comp));
  }

  /* Update local state. */
  cs = &comp->rstate[comp->recv_current = index];
  comp->flags &=~ SLF_TOSS;

  /* Save IP header (plus options). */
  memcpy(&cs->cs_ip, ip, sizeof(struct ip));		/* FIXME: 20 */
  if (ip->ip_hl > 5)
	memcpy(cs->cs_ipopt, ip+1, ((ip->ip_hl) - 5) * sizeof(unsigned long));

  /* Save TCP header (plus options). */
  memcpy(&cs->cs_tcp, thp, sizeof(struct tcphdr));	/* FIXME: 20 */
  if (thp->doff > 5)
	memcpy(cs->cs_tcpopt, thp+1, ((thp->doff) - 5) * sizeof(unsigned long));

  cs->cs_hsize = (ip->ip_hl * 2) + (thp->doff * 2);

  /*
   * Put headers back on packet.
   * Neither header checksum is recalculated.
   */
  comp->pp->stats.runcomp++;

  return(isize);
}
