/*
 * NET		An implementation of the SOCKET network access protocol.
 *		This module implements the PPP protocol for the kernel-
 *		based devices like TTY.  It interfaces between a raw TTY,
 *		and the kernel's NET protocol layers (via DDI).
 *
 *		Van Jacobsen Header Compression definitions.
 *
 * Version:	@(#)compress.h	1.7.0	04/02/94
 *
 * Authors:	Van Jacobson, <van@helios.ee.lbl.gov>, Dec 31, 1989
 *		See the `COPYRIGHT' file in this directory for the
 *		legal status and version info of this code.
 *
 *		Katie Stevens, <dkstevens@ucdavis.edu>, 01-31-90
 *		Phil Karn, <karn@ka9q.ampr.org>, 11-02-90
 *		Bill Simpson, <Bill_Simpson@um.cc.umich.edu>, 02-91
 *		Chuck Hedrick, <hedrick@klinzhai.rutgers.edu>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *		Jim Seagrave, <jes@grendel.demon.co.uk>
 *
 *		This program is  free  software; you  can redistribute  it
 *		and/or modify it under the terms of the GNU General Public
 *		License  as  published by the  Free  Software  Foundation;
 *		either version 2  of the License, or (at  your option) any
 *		later version.
 */
#include <linux/ip.h>
#include <linux/ip_tcp.h>


/* PPP compression masks for len/vers byte. */
#define PPP_TYPE_ERROR		0x00
#define PPP_TYPE_IP		0x40
#define PPP_TYPE_UNCOMP_TCP	0x70
#define PPP_TYPE_COMP_TCP	0x80

/* Bits in first octet of compressed packet. */
#define NEW_C		0x40			/* flag bits for what	*/
#define NEW_I		0x20			/* changed in a packet	*/
#define NEW_S		0x08
#define NEW_A		0x04
#define NEW_W		0x02
#define NEW_U		0x01

/* Reserved, special-case values of above. */
#define SPECIAL_I	(NEW_S|NEW_W|NEW_U)		/* echoed interactive traffic */
#define SPECIAL_D	(NEW_S|NEW_A|NEW_W|NEW_U)	/* unidirectional data */
#define SPECIALS_MASK	(NEW_S|NEW_A|NEW_W|NEW_U)

#define TCP_PUSH_BIT	0x10


#define SLF_TOSS	0x01			/* keep tossing until id rcv */


/* All the state data for one serial line (we need one of these per line). */
struct compress {
  struct cstate	*tstate;		/* transmit connection states	*/
  struct cstate *rstate;		/* receive connection states	*/

  unsigned char tslot_limit;		/* highest xmit slot id (0-l)	*/
  unsigned char rslot_limit;		/* highest recv slot id (0-l)	*/

  unsigned char	xmit_oldest;		/* oldest xmit in ring		*/
  unsigned char xmit_current;		/* most recent xmit id		*/
  unsigned char recv_current;		/* most recent rcvd id		*/

  unsigned char	flags;

  struct ppp	*pp;			/* back-pointer to PPP cb	*/
};


/*
 * "state" data for each active tcp conversation on the wire.  This is
 * basically a copy of the entire IP/TCP header from the last packet
 * we saw from the conversation together with a small identifier
 * the transmit & receive ends of the line use to locate saved header.
 */
struct cstate {
  unsigned char	connid;		/* connection id number (xmit)		*/
  struct cstate	*next;		/* next in ring (xmit)			*/
  struct ip	cs_ip;		/* IP hdr from most recent packet	*/
  struct tcphdr	cs_tcp;		/* TCP hdr from most recent packet	*/
  unsigned char	cs_ipopt[64];
  unsigned char	cs_tcpopt[64];
  int		cs_hsize;
};


extern struct compress	*pphc_init(struct ppp *pp, int rslots, int tslots);
extern void		pphc_free(struct compress *comp);
extern int		pphc_compress(struct compress *comp,
				      unsigned char *icp, int isize,
				      unsigned char *ocp,
				      unsigned char **cpp,
				      int compress_cid);
extern int		pphc_uncompress(struct compress *comp,
					unsigned char *icp, int isize);
extern int		pphc_remember(struct compress *comp,
				      unsigned char *icp, int isize);
extern int		pphc_toss(struct compress *comp);
extern void		pphc_print(unsigned char *ptr, int len);
extern void		pphc_i_status(struct compress *comp);
extern void		pphc_o_status(struct compress *comp);
