/*
 * NET		An implementation of the SOCKET network access protocol.
 *		This module implements the PPP protocol for the kernel-
 *		based devices like TTY.  It interfaces between a raw TTY,
 *		and the kernel's NET protocol layers (via DDI).
 *
 *		Define the PPP device driver internal functions.
 *
 * Version:	@(#)global.h	0.2.5	05/12/94
 *
 * Authors:	Michael Callahan <callahan@maths.ox.ac.uk>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *
 *		This program is  free  software; you  can redistribute  it
 *		and/or modify it under the terms of the GNU General Public
 *		License  as  published by the  Free  Software  Foundation;
 *		either version 2  of the License, or (at  your option) any
 *		later version.
 */
#ifndef _PPP_GLOBAL_H
#define _PPP_GLOBAL_H


#define OPTIMIZE_FLAG_TIME	((HZ * 3)/2)


#define ASSERT(p)		if (!p) DPRINTF(("assertion failed: " #p))



#define CHECK_PPP(a)		if (! pp->inuse) { \
					printk("PPP: *WARNING* using inactive interface in file %s, line %d !\n", \
						__FILE__, __LINE__); \
					return(a); \
				}
#define CHECK_PPP_VOID()	if (! pp->inuse) { \
					printk("PPP: *WARNING* using inactive interface in file %s, line %d !\n", \
						__FILE__, __LINE__); \
					return; \
				}

#define PUTC(c, label)		*pp->us_rbuff_head++ = c; \
				if (pp->us_rbuff_head == pp->us_rbuff_end) \
					pp->us_rbuff_head = pp->us_rbuff; \
				if (pp->us_rbuff_head == pp->us_rbuff_tail) \
					goto label;

#define GETC(c)			c = *pp->us_rbuff_tail++; \
				if (pp->us_rbuff_tail == pp->us_rbuff_end) \
					pp->us_rbuff_tail = pp->us_rbuff;

#define in_xmap(pp, c)		(pp->xmit_async_map[(c) >> 5] & \
							(1 << ((c) & 0x1f)))
#define in_rmap(pp, c)		((((unsigned int) (unsigned char) (c)) < 0x20) \
					&& pp->recv_async_map & (1 << (c)))

#define bset(p, b)		((p)[(b) >> 5] |= (1 << ((b) & 0x1f)))


struct ppp {
  int			magic;		/* magic value for structure	*/
  unsigned int		flags;		/* miscellany			*/
  char			inuse;		/* are we allocated?		*/
  char			sending;	/* "channel busy" indicator	*/
  char			escape;		/* 0x20 if prev char was PPP_ESC*/
  char			toss;		/* toss this frame		*/

  unsigned long		xmit_async_map[8];	/* 1 bit means that given
					 * control character is quoted
					 * on output			*/
  unsigned long		recv_async_map;		/* 1 bit means that given
					 * control character is ignored
					 * on input			*/
  int                   mtu;            /* maximum xmit frame size      */
  int                   mru;            /* maximum receive frame size   */
  unsigned short	fcs;		/* FCS field of current frame	*/

  unsigned long		mem_start,	/* memory pointers		*/
			mem_end,
			rmem_start,
			rmem_end;

  /* Various fields. */
  int			line;		/* PPP channel number		*/
  struct tty_struct	*tty;		/* ptr to TTY structure		*/
  struct ifnet		*dev;		/* easy for intr handling	*/
  unsigned long		last_xmit;	/* time of last transmission	*/

  /* VJ Header Compression ("CSLIP") control block. */
  struct compress	*comp;

  /* These are pointers to the malloc()ed frame buffers.
     These buffers are used while processing a packet.	If a packet
     has to hang around for the user process to read it, it lingers in
     the user buffers below. */
  unsigned char		*rbuff;		/* receiver buffer		*/
  unsigned char		*xbuff;		/* transmitter buffer		*/
  unsigned char		*cbuff;		/* compression buffer		*/

  /* These are the various pointers into the buffers. */
  unsigned char		*rhead;		/* RECV buffer pointer (head)	*/
  unsigned char		*rend;		/* RECV buffer pointer (end)	*/
  int			rcount;		/* PPP receive counter		*/
  unsigned char		*xhead;		/* XMIT buffer pointer (head)	*/
  unsigned char		*xtail;		/* XMIT buffer pointer (end)	*/

  /* Structures for interfacing with the user process. */
#define RBUFSIZE	4000
  unsigned char		*us_rbuff;	/* circular incoming packet buf.*/
  unsigned char		*us_rbuff_end;	/* end of allocated space	*/
  unsigned char		*us_rbuff_head; /* head of waiting packets	*/
  unsigned char		*us_rbuff_tail; /* tail of waiting packets	*/
  unsigned char		us_rbuff_lock;	/* lock: bit 0 head bit 1 tail	*/
  int			inp_sig;	/* input ready signal for pgrp	*/
  int			inp_sig_pid;	/* process to get notified	*/

  /* items to support the select() function */
  struct wait_queue	*write_wait;	/* queue for reading processes	*/
  struct wait_queue	*read_wait;	/* queue for writing processes	*/

  /* PPP interface statistics. */
  struct ppp_stats	stats;		/* statistic information	*/

  /* PPP demand dial information. */
  struct ppp_ddinfo	ddinfo;		/* demand dial information	*/
};


extern int		ppp_debug;	/* PPP debug level flag		*/
extern int		ppp_debug_netpackets;
#if CHECK_CHARS
extern unsigned long	paritytab;
#endif
extern struct ppp	*ppp_ctrl[];	/* PPP control blocks		*/
extern unsigned short	ppp_fcstab[];	/* PPP FCS table		*/


/* Fcs.c */
extern void		ppp_add_fcs(struct ppp *);
extern int		ppp_check_fcs(struct ppp *);

/* Main.c */
extern void		pp_stuff_char(struct ppp *pp, unsigned char c);
extern void		pp_enqueue(struct ppp *pp, unsigned char c);
extern int		pp_us_queue(struct ppp *pp, unsigned short proto,
				    unsigned char *buf, int len);
extern void		pp_changedmtu(struct ppp *, int new_mtu, int new_mru);
extern struct ppp	*pp_find(struct tty_struct *tty);
extern int		pp_open(struct ifnet *);
extern int		pp_close(struct ifnet *);
extern int		pp_lock(struct ppp *);
extern void		pp_unlock(struct ppp *);
extern void		pp_release(struct ppp *);

/* Tty.c */
#if NEW_TTY_DRIVERS
extern void		ppp_receive_buf(struct tty_struct *tty,
					unsigned char *cp, char *fp,
					int count);
#else
extern void		ppp_handler(struct tty_struct *);
#endif
extern void		ppp_kick_tty(struct ppp *);
extern void		ppp_close(struct tty_struct *);
extern int		ppp_open(struct tty_struct *);
extern int		ppp_ioctl(struct inode *inode, struct file *file,
				   unsigned int cmd, unsigned long arg);
extern int		ppp_read(struct tty_struct *tty, struct file *file,
				 unsigned char *buf, unsigned int nr);
extern int		ppp_write(struct tty_struct *tty, struct file *file,
				  unsigned char *buf, unsigned int nr);

extern void		ppp_write_wakeup(struct tty_struct *tty);
extern int		ppp_select(struct tty_struct *tty, struct inode *inode,
				   struct file *filp, int sel_type,
				   select_table * wait);

/* Packet.c */
extern void		ppp_output_done(void *pp);
extern void		ppp_doframe(struct ppp *);
extern int		ppp_xmit(struct ifnet *dev, struct sk_buff *skb);
extern int		ppp_recv(struct ifnet *dev, struct sk_buff *skb,
				  int len);
extern void		ppp_send(struct ifnet *dev, struct sk_buff *skb,
				  struct sockaddr *target, int priority);

/* Utils.c */
#ifdef PPP_DEBUG
extern void		ppp_dprintf(char *fmt, ...);
extern int		ppp_dbg_ioctl(unsigned long arg, int level);
extern void		ppp_print_buffer(const char *name, char *buf,
					int count, int seg);
#endif

#endif	/* _PPP_GLOBAL_H */
