/*
 * NET		An implementation of the SOCKET network access protocol.
 *		This module implements the PPP protocol for the kernel-
 *		based devices like TTY.  It interfaces between a raw TTY,
 *		and the kernel's NET protocol layers (via DDI).
 *
 *		PPP I/O control request handler.
 *
 * Version:	@(#)ioctl.h	0.2.5	05/12/94
 *
 * Authors:	Michael Callahan <callahan@maths.ox.ac.uk>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *
 *		This program is  free  software; you  can redistribute  it
 *		and/or modify it under the terms of the GNU General Public
 *		License  as  published by the  Free  Software  Foundation;
 *		either version 2  of the License, or (at  your option) any
 *		later version.
 */
#include <asm/segment.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/timer.h>
#include <linux/ddi.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/in_systm.h>
#include <linux/if.h>
#include <linux/if_ppp.h>
#include <linux/ip.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/tty.h>
#include "ppp.h"
#include "global.h"
#include "compress.h"


/* Check if the new MTU setting is OK. */
static int
ppp_check_mtu(struct ifnet *dev, int new_mtu)
{
  if ((new_mtu < 0) || (new_mtu >= 8000)) return(-EINVAL);
  dev->if_mtu = new_mtu;

  /* The driver will see the new MTU on the next write() ! */
  return(0);
}


/* Get or set the current PPP statistics. */
static int
ppp_do_stats(struct ppp *pp, unsigned long arg, unsigned cmd)
{
  int i;

  switch(cmd) {
	case SIOCGIFSTATS:
		i = verify_area(VERIFY_WRITE, (char *) arg,
				sizeof(struct ppp_stats));
		if (i < 0) return(i);
		memcpy_tofs((char *) arg, &pp->stats,
			    sizeof(struct ppp_stats));
		return(0);

	case SIOCSIFSTATS:
		i = verify_area(VERIFY_READ, (char *) arg,
				sizeof(struct ppp_stats));
		if (i < 0) return(i);
		memcpy_fromfs(&pp->stats, (char *) arg,
			    sizeof(struct ppp_stats));
		return(0);

	default:
		break;
  }
  return(-EINVAL);
}


/* Perform I/O control on an active PPP channel. */
int
ppp_ioctl(struct inode *inode, struct file *file,
	  unsigned int cmd, unsigned long arg)
{
  struct ppp_ddinfo cur_ddinfo;
  unsigned long temp_tbl[8];
  unsigned long cur_jiffies;
  register int i = 0;
  struct ifnet *dev;
  struct ppp *pp;
  int err;

  DPRINTF(("PPP: ioctl(0x%lX, 0x%lX, 0x%lX)\n", inode, cmd, arg));

#if NEW_TTY_DRIVER
  /*
   * These events must be passed to the real serial driver for them to
   * operate. Return the magic error condition to allow this to happen.
   */
  if (cmd == TIOCGETD ||
      cmd == TIOCSETD ||
      cmd == TCGETS   ||  /* allow stty to read, but not set, */
      cmd == TCGETA)      /* the serial port. */
	return(-ENOIOCTLCMD);
#endif

  switch(cmd) {
	case DDIOCSDBG:
		if (! suser()) return(-EPERM);
		err = verify_area(VERIFY_READ, (int *) arg, sizeof(int));
		if (err < 0) return(err);
		return(ppp_dbg_ioctl(arg, 1));

	case SIOCSIFMTU:
		dev = (struct ifnet *) inode;
		if (dev->if_unit >= PPP_MAX_DEV) return(-ENODEV);
		if (ppp_ctrl[dev->if_unit] == NULL) return(-ENODEV);
		return(ppp_check_mtu(dev, (int) arg));

	case SIOCGIFENCAP:		/* "get ENCAP mode"	*/
		return(-EINVAL);

	case SIOCSIFENCAP:		/* "set ENCAP mode"	*/
		if (! suser()) return(-EPERM);
		return(-EINVAL);

	case SIOCSIFSTATS:		/* "set stats" */
		if (! suser()) return(-EPERM);
		/*FALLTHROUGH*/

	case PPPIOCSDEBUG:
		err = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		ppp_debug = (int) get_fs_long(arg);
		ppp_debug_netpackets = (ppp_debug & 0xff00) >> 8;
		ppp_debug &= 0xff;
		DPRINTF(("PPP: ioctl: set debug level %d, netpacket %d\n", 
					ppp_debug, ppp_debug_netpackets));
		return(0);

	case PPPIOCGDEBUG:
		err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		put_fs_long((long) (ppp_debug | (ppp_debug_netpackets << 8)),
									arg);
		DPRINTF(("PPP: ioctl: get debug level %d\n", 
				ppp_debug | (ppp_debug_netpackets << 8)));
		return(0);

	case SIOCGIFNAME:
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		err = verify_area(VERIFY_WRITE, (int *) arg,
				(strlen(pp->dev->if_name) + 1));
		if (err < 0) return(err);
		memcpy_tofs((int *) arg, pp->dev->if_name,
			    strlen(pp->dev->if_name) + 1);
		return(0);

	case PPPIOCSMRU:
		err = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		i = (int) get_fs_long(arg);
		DPRINTF(("PPP: ioctl: set mru to %d\n", i));
		if (pp->mru != i) pp_changedmtu(pp, pp->mtu, i);
		return(0);

	case SIOCGIFSTATS:		/* "get stats" */
	case PPPIOCGSTAT:
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		return(ppp_do_stats(pp, arg, cmd));

	case PPPIOCGFLAGS:
		err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		i = (pp->flags & SC_MASK);
#if (CHECK_CHARS == 0)
		/* Don't generate errors if we don't check chars. */
		i |= (SC_RCV_B7_1 | SC_RCV_B7_0 | SC_RCV_ODDP | SC_RCV_EVNP);
#endif
		put_fs_long((long) i, arg);
		DPRINTF(("PPP: ioctl: get flags: addr %lX flags %lX\n",
				(unsigned long) arg, (unsigned long) i));
		return(0);

	case PPPIOCSFLAGS:
		err = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		i = (int) get_fs_long(arg);
		pp->flags ^= ((pp->flags ^ i) & SC_MASK);
		DPRINTF(("PPP: ioctl: set flags to %lX\n", (unsigned long) i));
		return(0);

	case PPPIOCGASYNCMAP:
		err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		put_fs_long(pp->xmit_async_map[0], arg);
		DPRINTF(("PPP: ioctl: get asyncmap: addr %lX asyncmap %lX\n",
						arg, pp->xmit_async_map[0]));
		return(0);

	case PPPIOCSASYNCMAP:
		err = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		memset(pp->xmit_async_map, 0, sizeof(pp->xmit_async_map));
		pp->xmit_async_map[0] = get_fs_long(arg);
		bset(pp->xmit_async_map, PPP_FLAG);
		bset(pp->xmit_async_map, PPP_ESC);
		DPRINTF(("PPP: ioctl: set asyncmap %lX\n",
					pp->xmit_async_map[0]));
		return(0);

	case PPPIOCRASYNCMAP:
		err = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		pp->recv_async_map = get_fs_long(arg);
		DPRINTF(("PPP: ioctl: set recv asyncmap %lX\n",
					pp->recv_async_map));
		return(0);
  
	case PPPIOCGUNIT:
		err = verify_area(VERIFY_WRITE, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		put_fs_long((long) pp->line, arg);
		DPRINTF(("PPP: ioctl: get unit: %d\n", pp->line));
		return(0);

	case PPPIOCSINPSIG:
		err = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		pp->inp_sig = (int) get_fs_long(arg);
		pp->inp_sig_pid = current->pid;
		DPRINTF(("PPP: ioctl: set input signal %d\n", pp->inp_sig));
		return(0);

	case PPPIOCGTIME:
		err = verify_area(VERIFY_WRITE, (void *) arg,
				    sizeof(struct ppp_ddinfo));
		if (err < 0) return(err);
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		cur_jiffies = jiffies;

		/* Change absolute times to relative times. */
		cur_ddinfo.ip_sjiffies =
			cur_jiffies - pp->ddinfo.ip_sjiffies;
		cur_ddinfo.ip_rjiffies =
			cur_jiffies - pp->ddinfo.ip_rjiffies;
		cur_ddinfo.nip_sjiffies =
			cur_jiffies - pp->ddinfo.nip_sjiffies;
		cur_ddinfo.nip_rjiffies =
			cur_jiffies - pp->ddinfo.nip_rjiffies;
    
		memcpy_tofs((void *) arg, &cur_ddinfo,
			    sizeof(struct ppp_ddinfo));
		DPRINTF(("PPP: ioctl: read demand dial info\n"));
		return(0);

	case PPPIOCGXASYNCMAP:
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		err = verify_area(VERIFY_WRITE, (void *) arg,
				  sizeof(pp->xmit_async_map));
		if (err < 0) return(err);
		memcpy_tofs((void *) arg, pp->xmit_async_map,
			    sizeof(pp->xmit_async_map));
		DPRINTF(("PPP: ioctl: get xasyncmap: addr %lX\n", arg));
		return(0);

	case PPPIOCSXASYNCMAP:
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		err = verify_area(VERIFY_READ, (void *) arg,
				  sizeof(pp->xmit_async_map));
		if (err < 0) return(err);
		memcpy_fromfs(temp_tbl, (void *) arg,
			      sizeof(pp->xmit_async_map));
		temp_tbl[1] = 0x00000000;	/* must not escape 0x20-0x3f */
		temp_tbl[2] &= ~0x40000000;	/* must not escape 0x5e	*/
		temp_tbl[3] |= 0x60000000;	/* must escape 0x7d and 0x7e */

		if ((temp_tbl[2] & temp_tbl[3]) != 0 ||
		    (temp_tbl[4] & temp_tbl[5]) != 0 ||
		    (temp_tbl[6] & temp_tbl[7]) != 0) return(-EINVAL);
		memcpy(pp->xmit_async_map, temp_tbl,
		       sizeof(pp->xmit_async_map));
		DPRINTF(("PPP: ioctl: set xasyncmap\n"));
		return(0);

	case PPPIOCSMAXCID:
		if ((pp = pp_find((struct tty_struct *) inode)) == NULL)
								return(-EINVAL);
		err = verify_area(VERIFY_READ, (void *) arg, sizeof(int));
		if (err < 0) return(err);
		i = (int) get_fs_long(arg) + 1;
		DPRINTF(("PPP: ioctl: set maxcid to %d\n", i));
		if (pp->comp != NULL) pphc_free(pp->comp);
		pp->comp = pphc_init(pp, i, i);
		if (pp->comp == NULL) {
			DPRINTF(("PPP: no space for compression buffers!\n"));
			pp_release(pp);
			return(-ENOMEM);
		}
		return(0);

	default:
		break;
  }

  return(-EINVAL);
}
