/*
 * NET		An implementation of the SOCKET network access protocol.
 *		This module implements the PPP protocol for the kernel-
 *		based devices like TTY.  It interfaces between a raw TTY,
 *		and the kernel's NET protocol layers (via DDI).
 *
 *		Main program of the PPP device driver.
 *
 * Version:	@(#)main.c	0.2.5	05/12/94
 *
 * Authors:	Michael Callahan <callahan@maths.ox.ac.uk>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *
 *		This program is  free  software; you  can redistribute  it
 *		and/or modify it under the terms of the GNU General Public
 *		License  as  published by the  Free  Software  Foundation;
 *		either version 2  of the License, or (at  your option) any
 *		later version.
 */
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/timer.h>
#include <linux/ddi.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/in_systm.h>
#include <linux/if.h>
#include <linux/if_ppp.h>
#include <linux/ip.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/tty.h>
#include "ppp.h"
#include "global.h"
#include "compress.h"


#define VERSION		"0.2.5"


static struct tty_ldisc	ppp_ldisc;
struct ppp		*ppp_ctrl[PPP_MAX_DEV];


/* Initialize a PPP control block for use. */
static void
pp_initialize(struct ppp *pp)
{
  pp->magic = PPP_MAGIC;
  pp->sending = 0;
  pp->toss = 0xFE;
  pp->escape = 0;

  pp->flags = 0;
  pp->mtu = PPP_MTU;
  pp->mru = PPP_MRU;
  pp->fcs = 0;

  memset(pp->xmit_async_map, 0, sizeof(pp->xmit_async_map));
  pp->xmit_async_map[0] = 0xffffffff;
  pp->xmit_async_map[3] = 0x60000000;
  pp->recv_async_map = 0x00000000;

  pp->comp = NULL;
  pp->rbuff = NULL;
  pp->xbuff = NULL;
  pp->cbuff = NULL;

  pp->rhead = NULL;
  pp->rend = NULL;
  pp->rcount = 0;
  pp->xhead = NULL;
  pp->xtail = NULL;

  pp->us_rbuff = NULL;
  pp->us_rbuff_end = NULL;
  pp->us_rbuff_head = NULL;
  pp->us_rbuff_tail = NULL;
  pp->read_wait = NULL;
  pp->write_wait = NULL;
  pp->us_rbuff_lock = 0;

  pp->inp_sig = 0;
  pp->inp_sig_pid = 0;

#ifdef OPTIMIZE_FLAG_TIME /* ensure flag will always be sent first time */
  pp->last_xmit	= jiffies - OPTIMIZE_FLAG_TIME;
#else
  pp->last_xmit	= 0;
#endif

  /* Clear statistics. */
  memset(&pp->stats, 0, sizeof(struct ppp_stats));

  /* Reset the demand dial information. */
  pp->ddinfo.ip_sjiffies  =
  pp->ddinfo.ip_rjiffies  =
  pp->ddinfo.nip_sjiffies =
  pp->ddinfo.nip_rjiffies = jiffies;
}


/*
 * Stuff a character into the transmit buffer, using PPP's way of
 * escaping special characters.  Slso, update pp->fcs to take
 * account of new character.
 */
inline void
pp_stuff_char(struct ppp *pp, unsigned char c)
{
  int curpt = pp->xhead - pp->xbuff;

  if ((curpt < 0) || (curpt > 3000)) {
	DPRINTF(("PPP: stuff_char: %lX %lX %d\n",
		(unsigned long) pp->xbuff, (unsigned long) pp->xhead, curpt));
  }
  if (in_xmap(pp, c)) {
	*pp->xhead++ = PPP_ESC;
	*pp->xhead++ = c ^ PPP_TRANS;
  } else {
	*pp->xhead++ = c;
  }
  pp->fcs = (pp->fcs >> 8) ^ ppp_fcstab[(pp->fcs ^ c) & 0xff];
}


inline void
pp_enqueue(struct ppp *pp, unsigned char c)
{
  unsigned long flags;

  save_flags(flags);
  cli();
  if (pp->rhead < pp->rend) {
	*pp->rhead = c;
	pp->rhead++;
	pp->rcount++;
  } else {
	pp->stats.roverrun++;
  }
  restore_flags(flags);
}


int
pp_us_queue(struct ppp *pp, unsigned short proto, unsigned char *buf, int len)
{
  unsigned char *saved_head;
  int totlen;

  totlen = (len + 2);		/* including protocol ID */

  if (set_bit(1, &pp->us_rbuff_lock)) {
	DPRINTF(("PPP: us_queue: can't get lock\n"));
	return(0);
  }
  saved_head = pp->us_rbuff_head;

  /* Set up the packet header. */
  PUTC((totlen & 0xff00) >> 8, failure);
  PUTC(totlen & 0x00ff, failure);
  PUTC((proto & 0xff00) >> 8, failure);
  PUTC(proto & 0x00ff, failure);

  /* Copy the packet data into the buffer. */
  while (len-- > 0) {
	PUTC(*buf++, failure);
  }

  DPRINTF(("PPP: successfully queued %u bytes\n", totlen));
  clear_bit(1, &pp->us_rbuff_lock);
  wake_up_interruptible(&pp->read_wait);

#if NEW_TTY_DRIVERS
  kill_fasync(pp->tty->fasync, SIGIO);
#endif

 if (pp->inp_sig && pp->inp_sig_pid) {
	if (kill_proc(pp->inp_sig_pid, pp->inp_sig, 1) != 0) {
		/* process is gone */
		DPRINTF(("PPP: process that requested notification is gone\n"));
		pp->inp_sig = 0;
		pp->inp_sig_pid = 0;
	}
	return(1);
  }

failure:
  pp->us_rbuff_head = saved_head;
  clear_bit(1, &pp->us_rbuff_lock);

  DPRINTF(("PPP: us_queue: ran out of buffer space.\n"));
  return(0);
}


/*
 * Marks a PPP interface 'busy'.  User processes will wait,
 * if they try to write, and the network code will refrain
 * from sending return nonzero if succeeded in acquiring lock.
 */
int
pp_lock(struct ppp *pp)
{
  int flags, locked;

  save_flags(flags);
  cli();
  locked = pp->sending;
  pp->sending = 1;
  if (pp->dev->if_flags & IFF_UP) pp->dev->if_tbusy = 1;
  restore_flags(flags);

  return(locked == 0);
}


void
pp_unlock(struct ppp *pp)
{
  int flags;

  save_flags(flags);
  cli();
  pp->sending = 0;
  if (pp->dev->if_flags & IFF_UP) pp->dev->if_tbusy = 0;
  restore_flags(flags);
}


/* Find a PPP channel from its `tty' link. */
struct ppp *
pp_find(struct tty_struct *tty)
{
  int i;

  if (tty == NULL) return(NULL);
  for (i = 0; i < PPP_MAX_DEV; i++) {
	if (ppp_ctrl[i] == NULL) continue;
	if (ppp_ctrl[i]->tty == tty) return(ppp_ctrl[i]);
  }
  return(NULL);
}


/*
 * MTU has been changed by the IP layer. Unfortunately we are not told
 * about this, but we spot it ourselves and fix things up. We could be
 * in an upcall from the tty driver, or in an IP packet queue.
 */
void
pp_changedmtu(struct ppp *pp, int new_mtu, int new_mru)
{
  unsigned char *new_rbuff, *new_xbuff, *new_cbuff;
  unsigned char *old_rbuff, *old_xbuff, *old_cbuff;
  struct ifnet *dev;
  int mtu, mru;

  /* Allocate the buffer from the kernel for the data. */
  dev = pp->dev;
  mru = new_mru;
  mtu = new_mtu;

  /* RFC 1331, section 7.2 says the minimum value is 1500 bytes */
  if (mru < PPP_MRU) mru = PPP_MRU;
  mtu = (mtu * 2) + 20;
  mru = (mru * 2) + 20;

  DPRINTF(("PPP: channel %s mtu=%d mru=%d\n", dev->if_name, new_mtu, new_mru));
  new_xbuff = (unsigned char *) kmalloc(mtu + 4, GFP_KERNEL);
  new_rbuff = (unsigned char *) kmalloc(mru + 4, GFP_KERNEL);
  new_cbuff = (unsigned char *) kmalloc(mru + 4, GFP_KERNEL);

  /* If the buffers failed to allocate then complain. */
  if (new_xbuff == NULL || new_rbuff == NULL || new_cbuff == NULL) {
	DPRINTF(("PPP: failed to allocate new buffers\n"));

	/* Release new buffer pointers if the updates were not performed. */
	if (new_rbuff != NULL) kfree(new_rbuff);
	if (new_xbuff != NULL) kfree(new_xbuff);
	if (new_cbuff != NULL) kfree(new_cbuff);
  } else {
	/* Update the pointers to the new buffer structures. */
	cli();
	old_xbuff = pp->xbuff;
	old_rbuff = pp->rbuff;
	old_cbuff = pp->cbuff;

	pp->xbuff = new_xbuff;
	pp->rbuff = new_rbuff;
	pp->cbuff = new_cbuff;

	pp->mem_start = (unsigned long) new_xbuff;
	pp->mem_end = (unsigned long) (pp->mem_start + mtu);

	pp->rmem_start = (unsigned long) new_rbuff;
	pp->rend = (unsigned char *)
	pp->rmem_end = (unsigned long) (pp->rmem_start + mru);

	pp->rhead = new_rbuff;

	/* Update the parameters for the new buffer sizes. */
	pp->toss = 0xFE;
	pp->escape = 0;
	pp->sending = 0;
	pp->rcount = 0;
	pp->mtu = dev->if_mtu = new_mtu;
	pp->mru = new_mru;

	sti();

	/* Release old buffer pointers. */
	if (old_rbuff != NULL) kfree(old_rbuff);
	if (old_xbuff != NULL) kfree(old_xbuff);
	if (old_cbuff != NULL) kfree(old_cbuff);
  }
}


/* Find a free PPP channel, and link in this `tty' line. */
static struct ppp *
pp_alloc(void)
{
  int i;

  for (i = 0; i < PPP_MAX_DEV; i++) {
	if (ppp_ctrl[i] == NULL) break;
	if (! set_bit(0, &ppp_ctrl[i]->inuse)) {

		/* UPping the device here sounds fishy to me -FvK */
		ppp_ctrl[i]->dev->if_flags |= (IFF_UP | IFF_RUNNING);
		ppp_ctrl[i]->tty = NULL;
		return(ppp_ctrl[i]);
	}
  }
  return(NULL);
}


/* Called when we abandon the PPP line discipline */
void
pp_release(struct ppp *pp)
{
#if NEW_TTY_DRIVERS
  if (pp->tty != NULL && pp->tty->disc_data == ppp)
		pp->tty->disc_data = NULL; /* Break the tty->ppp link */
#endif

  kfree(pp->xbuff);
  kfree(pp->cbuff);
  kfree(pp->rbuff);
  kfree(pp->us_rbuff);

  pp->xbuff = NULL;
  pp->cbuff = NULL;
  pp->rbuff = NULL;
  pp->us_rbuff = NULL;

  if (pp->comp) {
	pphc_free(pp->comp);
	pp->comp = NULL;
  }

  /* Close down the NET interface. */
  (void) dev_close(pp->dev);
  pp->dev->if_flags |= IFF_POINTOPOINT;

  pp->inuse = 0;
  pp->tty = NULL;
}


/*
 * Called when PPP interface goes "up".
 * Here this just means we start passing IP packets. */
int
pp_open(struct ifnet *dev)
{
  struct ppp *pp = ppp_ctrl[dev->if_unit];

  if (pp->tty == NULL) {
	DPRINTF(("PPP: %s not connected to a TTY! can't go open!\n",
							dev->if_name));
	return(-ENXIO);
  }

  /* Reset POINTOPOINT every time, since dev_close zaps it! */
  dev->if_flags |= IFF_POINTOPOINT;

  DPRINTF(("PPP: channel %s going up for IP packets!\n", dev->if_name));

  CHECK_PPP(-ENXIO);
  return(0);
}


int
pp_close(struct ifnet *dev)
{
  struct ppp *pp = ppp_ctrl[dev->if_unit];

  if (pp->tty == NULL) {
	DPRINTF(("PPP: %s not connected to a TTY! can't go down!\n",
							dev->if_name));
	return(-ENXIO);
  }

  DPRINTF(("PPP: channel %s going down for IP packets!\n", dev->if_name));

  CHECK_PPP(-ENXIO);
  return(0);
}


/* Called when PPP line discipline is selected on a tty. */
int
ppp_open(struct tty_struct *tty)
{
  struct ppp *pp = pp_find(tty);

  if (pp != NULL) {
	DPRINTF(("PPP: open: gack! tty already associated to %s!\n",
		pp->magic == PPP_MAGIC ? pp->dev->if_name : "unknown"));
		return(-EEXIST);
  }

  pp = pp_alloc();
  if (pp == NULL) {
	DPRINTF(("PPP: open: couldn't allocate ppp channel\n"));
	return(-ENFILE);
  }

  /* Make sure the channel is actually open. */
  pp_initialize(pp);

  pp->tty = tty;

#if NEW_TTY_DRIVERS
  tty->disc_data = pp;
  if (tty->driver.flush_buffer) tty->driver.flush_buffer(tty);
  if (tty->ldisc.flush_buffer) tty->ldisc.flush_buffer(tty);
#else
  tty_read_flush(tty);
  tty_write_flush(tty);
#endif

  if ((pp->comp = pphc_init(pp, 16, 16)) == NULL) {
	DPRINTF(("PPP: no space for compression buffers!\n"));
	pp_release(pp);
	return(-ENOMEM);
  }

  /* Define the buffers for operation. */
  pp_changedmtu(pp, pp->dev->if_mtu, pp->mru);
  if (pp->rbuff == NULL) {
	pp_release(pp);
	return(-ENOMEM);
  }

  /* Allocate a user-level receive buffer. */
  pp->us_rbuff = kmalloc(RBUFSIZE, GFP_KERNEL);
  if (pp->us_rbuff == NULL) {
	DPRINTF(("PPP: no space for user receive buffer\n"));
	pp_release(pp);
	return(-ENOMEM);
  }

  pp->us_rbuff_head = pp->us_rbuff;
  pp->us_rbuff_tail = pp->us_rbuff;
  pp->us_rbuff_end = (pp->us_rbuff + RBUFSIZE);

  DPRINTF(("PPP: channel %s open\n", pp->dev->if_name));

#if NEW_TTY_DRIVERS
  return(0);
#else
  return(pp->line);
#endif
}


void
ppp_close(struct tty_struct *tty)
{
  struct ppp *pp = pp_find(tty);

  if (pp == NULL || pp->magic != PPP_MAGIC) {
	DPRINTF(("PPP: trying to close unopened tty!\n"));
  } else {
	CHECK_PPP_VOID();
	pp_release(pp);
	DPRINTF(("PPP: channel %s closing.\n", pp->dev->if_name));
  }
}


static int
ppp_fopen(struct inode *inode, struct file *file)
{
  int minor;

  minor = MINOR(inode->i_rdev);
  if (minor < 0 || minor >= PPP_MAX_DEV) return(-ENODEV);

  return(0);
}


static void
ppp_fclose(struct inode *inode, struct file *file)
{
  int minor;

  minor = MINOR(inode->i_rdev);
  if (minor < 0 || minor >= PPP_MAX_DEV) return;
}


static struct file_operations ppp_fops = {
  NULL,		/* LSEEK	*/
  NULL,		/* READ		*/
  NULL,		/* WRITE	*/
  NULL,		/* READDIR	*/
  NULL,		/* SELECT	*/
  ppp_ioctl,	/* IOCTL	*/
  NULL,		/* MMAP		*/
  ppp_fopen,	/* OPEN		*/
  ppp_fclose	/* CLOSE	*/
};


int
ppp_init(struct ddi *ddi)
{
  struct ifnet *dev;
  struct ppp *pp;
  int i;

  printk("%s: version %s", ddi->name, VERSION);
#if NEW_TTY_DRIVERS
  printk(" (NEW_TTY_DRV)");
#endif
  printk(": ");

  /* Check for sanity here... */
  if (ddi->nunits >= PPP_MAX_DEV) {
	printk("too many channels requested!\n");
	return(-1);
  }

  /* Register VFS device. */
  if (ddi->major != 0) {
	if (register_chrdev(ddi->major, ddi->name, &ppp_fops) < 0) {
		printk("cannot register device %d!\n", ddi->major);
		return(-EINVAL);
	}
	printk("device %d/%d: ", ddi->major, ddi->minor);
  }
  printk("%d channels: ", ddi->nunits);

  /* Fill in our LDISC request block. */
  (void) memset(&ppp_ldisc, 0, sizeof(ppp_ldisc));
  ppp_ldisc.flags = 0;
  ppp_ldisc.open = ppp_open;
  ppp_ldisc.close = ppp_close;
  ppp_ldisc.read = ppp_read;
  ppp_ldisc.write  = ppp_write;
  ppp_ldisc.select = ppp_select;
  ppp_ldisc.ioctl = (int (*)(struct tty_struct *, struct file *,
			   unsigned int, unsigned long)) ppp_ioctl;
#if NEW_TTY_DRIVERS
  ppp_ldisc.magic = TTY_LDISC_MAGIC;
  ppp_ldisc.receive_buf = ppp_receive_buf;
  ppp_ldisc.write_wakeup = ppp_write_wakeup;
#else
  ppp_ldisc.handler = ppp_handler;
#endif

  if (tty_register_ldisc(N_PPP, &ppp_ldisc) != 0) {
	printk("cannot register line discipline %d !\n", N_PPP);
	return(-ENODEV);
  }
  printk("OK\n");
  
  /* Set up all our channels. */
  for(i = 0; i < ddi->nunits; i++) {
	pp = (struct ppp *) kmalloc(sizeof(struct ppp), GFP_KERNEL);
	if (pp == NULL) {
		printk("PPP: no memory for control block #%d !\n", i);
		return(-ENOMEM);
	}
	memset(pp, 0, sizeof(struct ppp));

	if ((dev = dev_create(NULL)) == NULL) {
		printk("PPP: no memory for interface descriptor #%d !\n", i);
		kfree_s(pp, sizeof(struct ppp));
		return(-ENOMEM);
	}
	ppp_ctrl[i] = pp;
	pp->line = i;
	pp->dev = dev;
	pp_initialize(pp);
	pp->inuse = 0;
	pp->tty = NULL;

	sprintf(dev->if_name, PPP_IFNAME, i);
	dev->if_flags = IFF_POINTOPOINT;
	dev->if_mtu = PPP_MTU;
	dev->if_unit = pp->line;
	dev->if_tbusy = 0;
	dev->if_machlen	= PPP_HLEN;
	dev->if_macalen	= PPP_ALEN;
	dev->if_macaddr->ifa_addr.sa_family = ARPHRD_PPP;

	dev->mac1 = NULL;
	dev->mac2 = NULL;
	dev->mac_aget = NULL;
	dev->mac_aset = NULL;
	dev->mac_bget = NULL;
	dev->mac_bset = NULL;
	dev->mac_print = NULL;
	dev->mac_sprint = NULL;
	dev->set_multi = NULL;

	dev->open = pp_open;
	dev->close = pp_close;
	dev->ioctl = (int (*)(struct inode *, struct file *,
			   unsigned int, unsigned long)) ppp_ioctl;
	dev->xmit = ppp_xmit;
	dev->send = ppp_send;
	dev->recv = ppp_recv;
  }

  /* Clear out the remaining PPP channels. */
  while (i < PPP_MAX_DEV)
	ppp_ctrl[i++] = NULL;

  /* All done... */
  return(0);
}
