/*
 * NET		An implementation of the SOCKET network access protocol.
 *		This module implements the PPP protocol for the kernel-
 *		based devices like TTY.  It interfaces between a raw TTY,
 *		and the kernel's NET protocol layers (via DDI).
 *
 *		PPP packet handling functions.
 *
 * Version:	@(#)packet.c	0.2.5	05/12/94
 *
 * Authors:	Michael Callahan <callahan@maths.ox.ac.uk>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *
 *		This program is  free  software; you  can redistribute  it
 *		and/or modify it under the terms of the GNU General Public
 *		License  as  published by the  Free  Software  Foundation;
 *		either version 2  of the License, or (at  your option) any
 *		later version.
 */
#include <asm/segment.h>
#include <asm/bitops.h>
#include <asm/system.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/timer.h>
#include <linux/ddi.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/in_systm.h>
#include <linux/if.h>
#include <linux/if_ppp.h>
#include <linux/if_ether.h>		/* for ETH_P_IP */
#include <linux/ip.h>
#include <linux/tty.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/tty.h>
#include "ppp.h"
#include "global.h"
#include "compress.h"


/*
 * Examine packet at C, attempt to pass up to net layer. 
 * PROTO is the protocol field from the PPP frame.
 * Return 1 if could handle it, 0 otherwise.
 */
static int
ppp_do_ip(struct ppp *pp, unsigned short proto, unsigned char *c, int count)
{
  struct sk_buff *skb;
  int flags, done;

  DPRINTF(("PPP: do_ip: proto %X len %u first byte %X\n",
				(int) proto, count, (int) c[0]));

  if (ppp_debug_netpackets) {
	DPRINTF((" %s <-- proto %X len %u\n",
		pp->dev->if_name, (int) proto, count));
  }
    
  if (proto == PROTO_IP) {
	pp->stats.runcomp++;
	goto sendit;
  }

  if ((proto == PROTO_VJCOMP) && !(pp->flags & SC_REJ_COMP_TCP)) {
	/* Get space for uncompressing the header. */
	done = 0;
	save_flags(flags);
	cli();
	if ((pp->rhead + 80) < pp->rend) {
		pp->rhead += 80;
		pp->rcount += 80;
		done = 1;
	}
	restore_flags(flags);

	if (! done) {
		DPRINTF(("PPP: no space to decompress VJ compressed TCP header.\n"));
		pp->stats.roverrun++;
		pp->dev->if_stats->rx_dropped++;
		return(1);
	}

	count = pphc_uncompress(pp->comp, c, count);
	if (count <= 0) {
		pp->stats.rerrors++;
		pp->dev->if_stats->rx_dropped++;
		DPRINTF(("PPP: error in VJ decompression\n"));
		return(1);
	}
	pp->stats.rcomp++;
	goto sendit;
  }
  
  if ((proto == PROTO_VJUNCOMP) && !(pp->flags & SC_REJ_COMP_TCP)) {
	if (pphc_remember(pp->comp, c, count) <= 0) {
		pp->stats.rerrors++;
		pp->dev->if_stats->rx_dropped++;
		DPRINTF(("PPP: error in VJ memorizing\n"));
		return(1);
	}
	pp->stats.runcomp++;
	goto sendit;
  }

  /* Not ours... */
  pp->dev->if_stats->rx_dropped++;
  return(0);

sendit:
#ifdef PPP_DEBUG
  if (ppp_debug_netpackets) {
	struct ip *iph = (struct ip *) c;

	DPRINTF(("%s <--    src %lX dst %lX len %u\n",
		pp->dev->if_name, iph->ip_src.s_addr,
		iph->ip_dst.s_addr, count));
  }
#endif

  /* Allocate skb and copy data. */
  if ((skb = skb_create(count)) == NULL) {
	printk("%s: Could not allocate skb of size %u !\n",
			pp->dev->if_name, count);
	pp->dev->if_stats->rx_dropped++;
	return(0);
  }

  /* Copy the packet to the new buffer. */
  memcpy(skb->raw, c, count);

  /* Bump the datagram to the upper layers... */
  DPRINTF(("PPP: packet is %u at 0x%lX\n", count, (unsigned long) skb));
  do {
        done = dev_rint(pp->dev, skb, count);
        if (done == 0 || done == 1) break;
  } while(1);

  /* Update the statistics. */
  pp->dev->if_stats->rx_packets++;
  if (done == 1) pp->dev->if_stats->rx_dropped++;

  return(1);
}


/*
 * On entry, a received frame is in pp->rbuff
 * check it and dispose as appropriate.
 */
void
ppp_doframe(struct ppp *pp)
{
  u_char *c = pp->rbuff;
  u_short proto;
  int count = pp->rcount;

  /* Forget it if we've already noticed an error. */
  if (pp->toss) {
	DPRINTF(("PPP: toss: tossing frame, reason = %d\n", pp->toss));
	pp->stats.rerrors++;
	pp->dev->if_stats->rx_dropped++;
	return;
  }

  /* Do this before printing buffer to avoid generating copious output. */
  if (count == 0) return;

#ifdef PPP_DEBUG
  if (ppp_debug >= 3)
	ppp_print_buffer("receive frame", c, count, KERNEL_DS);
#endif

  if (count < 4) {
	DPRINTF(("PPP: got runt PPP frame, %u chars\n", count));
	pp->stats.runts++;
	pp->dev->if_stats->rx_dropped++;
	return;
  }

  /* Check PPP error detection field. */
  if (! ppp_check_fcs(pp)) {
	DPRINTF(("PPP: frame with bad fcs\n"));
	pp->stats.rerrors++;
	pp->dev->if_stats->rx_dropped++;
	return;
  }
  count -= 2;		/* ignore last two characters (== FCS). */

  /*
   * Now we have a good frame.
   * Figure out the protocol field.
   */
  if ((c[0] == PPP_ADDRESS) && (c[1] == PPP_CONTROL)) {
	c += 2;			/* ADDR/CTRL not compressed, so skip */
	count -= 2;
  }

  proto = (u_short) *c++;		/* PROTO compressed */
  if (proto & 1) {
	count--;
  } else {
	proto = (proto << 8) | (u_short) *c++;	/* PROTO uncompressed */
	count -= 2;
  }

  /* Send the frame to the network if the PPP interface is up. */
  if (pp->dev->if_flags & IFF_UP) {
	if (ppp_do_ip(pp, proto, c, count)) {
		pp->ddinfo.ip_rjiffies = jiffies;
		return;
	}
  }

  /*
   * If we got here, it has to go to a user process doing a
   * read, so queue it.
   *
   * User process expects to get whole frame (for some reason),
   * so use count+2 so as to include FCS field.
   */
  if (pp_us_queue(pp, proto, c, count+2)) {
	pp->ddinfo.nip_rjiffies = jiffies;
	pp->dev->if_stats->rx_packets++;
	pp->stats.rothers++;
	return;
  }

  /* Couldn't cope. */
  DPRINTF(("PPP: dropping packet on the floor: nobody could take it.\n"));
  pp->stats.tossed++;
  pp->dev->if_stats->rx_dropped++;
}


/* We finished some buffer. */
void
ppp_output_done(void *pp)
{
  /* Unlock the transmitter queue. */
  pp_unlock((struct ppp *) pp);

  /*
   * If the device is still up then enable the transmitter
   * of the next frame.
   */
  if (((struct ppp *) pp)->dev->if_flags & IFF_UP)
	dev_tint(((struct ppp *) pp)->dev);

  /* Enable any blocked process pending transmission. */
  wake_up_interruptible(&((struct ppp *) pp)->write_wait);
}


/* Process an incoming packet. */
int
ppp_recv(struct ifnet *dev, struct sk_buff *skb, int len)
{
  DPRINTF(("PPP: %s recv: len=%d\n", dev->if_name, len));

  /* PPP does not have a MAC header. */
  skb->raw = (unsigned char *) (skb + 1);
  skb->len -= dev->if_machlen;

  return(ETH_P_IP);
}


/* Send out a packet. */
void
ppp_send(struct ifnet *dev, struct sk_buff *skb,
	  struct sockaddr *target, int priority)
{
  DPRINTF(("PPP: %s sent: len=%d\n", dev->if_name, skb->len));

  skb->raw = (unsigned char *) (skb + 1);
#if 0
  skb->len += dev->if_machlen;
#endif

  /* Call the DEV interface for the actual work. */
  dev_queue(dev, skb, target, priority);
}


/* Encapsulate a packet and kick it into a TTY queue. */
int
ppp_xmit(struct ifnet *dev, struct sk_buff *skb)
{
  struct tty_struct *tty;
  struct ppp *pp;
  unsigned char *p;
  unsigned short proto;
  int len;

  if (skb == NULL) {
	DPRINTF(("PPP: xmit: null skb !\n"));
	return(0);
  }

  /* Find the correct PPP channel to use. */
  pp = ppp_ctrl[dev->if_unit];
  tty = pp->tty;
  p = skb->raw;
  len = skb->len;
  proto = PROTO_IP;

  DPRINTF(("PPP: xmit [%s]: skb %lX busy %d\n",
		dev->if_name, (unsigned long) skb, pp->sending));

  CHECK_PPP(0);

  if (tty == NULL) {
	DPRINTF(("PPP: xmit: %s not connected to a TTY!\n", dev->if_name));
	pp->dev->if_stats->tx_dropped++;
	goto done;
  }

  if (! (dev->if_flags & IFF_UP)) {
	DPRINTF(("PPP: xmit: packet sent on interface %s, which is down for IP\n",
					dev->if_name));
	pp->dev->if_stats->tx_dropped++;
	goto done;
  }

  /* Don't believe what the skb says- get the frame size from the header. */
  if (len < sizeof(struct ip)) {
	DPRINTF(("PPP: xmit: given runt packet, ignoring\n"));
	pp->dev->if_stats->tx_dropped++;
	return(1);
  }
  len = ntohs(((struct ip *) skb->raw)->ip_len);

  /* If doing demand dial then divert the first frame to pppd. */
  if (pp->flags & SC_IP_DOWN) {
	if (pp->flags & SC_IP_FLUSH == 0) {
		if (pp_us_queue(pp, proto, p, len)) pp->flags |= SC_IP_FLUSH;
	}
#if 1	/* FIXME: not really, as it will only be delayed for a while... */
	pp->dev->if_stats->tx_dropped++;
#endif
	goto done;
  }

  /* Attempt to acquire send lock. */
  if (pp->sending || !pp_lock(pp)) {
	DPRINTF(("PPP: xmit: BUSY\n"));
	pp->stats.sbusy++;
	pp->dev->if_stats->tx_dropped++;
	return(1);
  }
  pp->xhead = pp->xbuff;

  /* Try to compress, if VJ compression mode is on. */
  if (pp->flags & SC_COMP_TCP) {
	/* NOTE: last 0 argument says never to compress connection ID. */
	len = pphc_compress(pp->comp, p, len, pp->cbuff, &p, 0);
	if (p[0] & PPP_TYPE_COMP_TCP) proto = PROTO_VJCOMP;
	  else {
		if (p[0] >= PPP_TYPE_UNCOMP_TCP) {
			proto = PROTO_VJUNCOMP;
			p[0] = (p[0] & 0x0f) | 0x40; 
		}
	}
  }

  /* Increment appropriate counter. */
  if (proto == PROTO_VJCOMP) ++pp->stats.scomp;
    else ++pp->stats.suncomp;
      
#ifdef PPP_DEBUG
  if (ppp_debug_netpackets) {
	struct ip *iph = (struct ip *) (skb + 1);

	DPRINTF(("%s ==> proto %X len %d src %lX dst %lX proto %d\n",
		dev->if_name, (int) proto, (int) len,
		(unsigned long) iph->ip_src.s_addr,
		(unsigned long) iph->ip_dst.s_addr,
		(int) iph->ip_p));
  }
#endif

  /* start of frame:   FLAG  ALL_STATIONS  CONTROL  <protohi> <protolo> */
#ifdef OPTIMIZE_FLAG_TIME
  if (jiffies - pp->last_xmit > OPTIMIZE_FLAG_TIME) *pp->xhead++ = PPP_FLAG;
  pp->last_xmit = jiffies;
#else      
  *pp->xhead++ = PPP_FLAG;
#endif

  pp->fcs = PPP_FCS_INIT;
  if (! (pp->flags & SC_COMP_AC)) { 
	pp_stuff_char(pp, PPP_ADDRESS);
	pp_stuff_char(pp, PPP_CONTROL);
  }

  if (! (pp->flags & SC_COMP_PROT) || (proto & 0xff00)) {
	pp_stuff_char(pp, proto >> 8);
  }
  pp_stuff_char(pp, proto & 0xff);

  /* Data part. */
  while (len-- > 0)
	pp_stuff_char(pp, *p++);

  /* RCS and flag. */
  ppp_add_fcs(pp);
  *pp->xhead++ = PPP_FLAG;

  /* Update the time for demand dial function. */
  pp->ddinfo.ip_sjiffies = jiffies;

  /* Send it! */
  DPRINTF(("PPP: xmit: sending %d bytes\n", pp->xhead - pp->xbuff));
#ifdef PPP_DEBUG
  if (ppp_debug >= 6)
	ppp_print_buffer("xmit buffer", pp->xbuff,
			 pp->xhead - pp->xbuff, KERNEL_DS);
#endif
  ppp_kick_tty(pp);
  pp->dev->if_stats->tx_packets++;

 done:
  return(0);
}
