/*
 * NET		An implementation of the SOCKET network access protocol.
 *		This module implements the PPP protocol for the kernel-
 *		based devices like TTY.  It interfaces between a raw TTY,
 *		and the kernel's NET protocol layers (via DDI).
 *
 *		The following functions form support user programs which
 *		read and write data on a TTY with the PPP line discipline.
 *		Reading is done from a circular queue, filled by the lower
 *		TTY levels.
 *
 * Version:	@(#)tty.c	0.2.5	05/12/94
 *
 * Authors:	Michael Callahan <callahan@maths.ox.ac.uk>
 *		Fred N. van Kempen, <waltje@uwalt.nl.mugnet.org>
 *
 *		This program is  free  software; you  can redistribute  it
 *		and/or modify it under the terms of the GNU General Public
 *		License  as  published by the  Free  Software  Foundation;
 *		either version 2  of the License, or (at  your option) any
 *		later version.
 */
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/interrupt.h>
#include <linux/string.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/tty.h>
#include <linux/signal.h>
#include <linux/ddi.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/skbuff.h>
#include <linux/in.h>
#include <linux/in_systm.h>
#include <linux/if.h>
#include <linux/if_ppp.h>
#include <linux/ip.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include "ppp.h"
#include "global.h"
#include "compress.h"


#if NEW_TTY_DRIVERS
void
ppp_kick_tty(struct ppp *pp)
{
  register int count, actual;

  count = pp->xhead - pp->xbuff;

  actual = pp->tty->driver.write(pp->tty, 0, pp->xbuff, count);
  pp->stats.sbytes += actual;
  if (actual == count) {
	ppp_output_done(pp);
  } else {
	pp->xtail = pp->xbuff + actual;
	pp->tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);
  }
}


void
ppp_write_wakeup(struct tty_struct *tty)
{
  register int count, actual;
  struct ppp *pp = pp_find(tty);

  if (ppp == NULL || pp->magic != PPP_MAGIC) {
 	printk("PPP: write_wakeup called but couldn't find PPP struct.\n");
	return;
  }

  if (pp->xtail == NULL || (pp->flags & SC_XMIT_BUSY)) return;

  cli();
  if (pp->flags & SC_XMIT_BUSY) return;
  pp->flags |= SC_XMIT_BUSY;
  sti();
          
  count = pp->xhead - pp->xtail;
  actual = tty->driver.write(tty, 0, pp->xtail, count);
  pp->stats.sbytes += actual;
  if (actual == count) {
	pp->xtail = NULL;
	tty->flags &= ~TTY_DO_WRITE_WAKEUP;
	ppp_output_done(pp);
  } else {
	pp->xtail += actual;
  }
  pp->flags &= ~SC_XMIT_BUSY;
}


void
ppp_receive_buf(struct tty_struct *tty, unsigned char *cp,
		char *fp, int count)
{
  register struct ppp *pp = pp_find(tty);
  unsigned char c;
 
  if (pp == NULL || pp->magic != PPP_MAGIC) {
	DPRINTF(("PPP: handler called but couldn't find PPP struct.\n"));
	return;
  }

  CHECK_PPP_VOID();

#ifdef PPP_DEBUG
  if (ppp_debug >= 5)
	ppp_print_buffer("receive buffer", cp, count, KERNEL_DS);
#endif
 
  while (count-- > 0) {
	c = *cp++;

	if (fp) {
		if (*fp && pp->toss == 0) pp->toss = *fp;
		fp++;
	}

#if CHECK_CHARS
	if (c & 0x80)
		pp->flags |= SC_RCV_B7_1;
	  else
		pp->flags |= SC_RCV_B7_0;
	if (paritytab[c >> 5] & (1 << (c & 0x1F)))
		pp->flags |= SC_RCV_ODDP;
	  else
		pp->flags |= SC_RCV_EVNP;
#endif

	switch(c) {
		case PPP_ESC:	/* PPP_ESC: invert 0x20 in next character */
			pp->escape = PPP_TRANS;
			break;

		case PPP_FLAG:	/* PPP_FLAG: end of frame */
			/* PPP_ESC just before PPP_FLAG is "cancel". */
			if (pp->escape) pp->toss = 0xFF;

			/* Pass frame on to next layers. */
			if (! (pp->toss & 0x80)) ppp_doframe(pp);

			/* State Reset. */
			pp->rcount = 0;
			pp->rhead = pp->rbuff;
			pp->escape = 0;
			pp->toss = 0;
			break;

		default:	/* regular character */
			if (! in_rmap(pp, c)) {
				if (pp->toss == 0)
					ppp_enqueue(pp, c ^ pp->escape);
				pp->escape = 0;
			}
	}
  }
}
#else	/* ! NEW_TTY_DRIVERS */
#ifdef PPP_DEBUG
static void
ppp_dump_inqueue(struct tty_struct *tty)
{
  int  head = tty->read_q.head,
       tail = tty->read_q.tail,
       i, count;
  char buffer[8];

  DPRINTF(("INQUEUE: head %lX tail %lX imode %X:\n",
	(unsigned long) head, (unsigned long) tail, 
	(unsigned int) tty->termios->c_iflag));

  i = tail;
  count = 0;
  while (i != head) {
	buffer[count] = tty->read_q.buf[i];
	if (++count == 8) {
		ppp_print_buffer(NULL, buffer, 8, KERNEL_DS);
		count = 0;
	}
	i = (i + 1) & (TTY_BUF_SIZE - 1);
  }
  ppp_print_buffer(NULL, buffer, count, KERNEL_DS);
}
#endif	/* PPP_DEBUG */


/*
 * recover frame by undoing PPP escape mechanism;
 * copies N chars of input data from C into PPP->rbuff
 * calls ppp_doframe to dispose of any frames it finds
 */
static void
ppp_unesc(struct ppp *pp, unsigned char *c, int n)
{
  int i;

  for (i = 0; i < n; i++, c++) {
	DPRINTF(("PPP: (%X)", (unsigned int) *c));

#if CHECK_CHARS
	if (*c & 0x80)
		pp->flags |= SC_RCV_B7_1;
	  else
		pp->flags |= SC_RCV_B7_0;
	if (paritytab[*c >> 5] & (1 << (*c & 0x1F)))
		pp->flags |= SC_RCV_ODDP;
	  else
		pp->flags |= SC_RCV_EVNP;
#endif

	switch(*c) {
		case PPP_ESC:		/* PPP_ESC: invert 0x20 in next character */
			pp->escape = PPP_TRANS;
			break;

		case PPP_FLAG:		/* PPP_FLAG: end of frame */
			/* PPP_ESC just before PPP_FLAG is illegal. */
			if (pp->escape) pp->toss = 0xFF;

			/* Pass frame on to next layers. */
			if (! (pp->toss & 0x80)) ppp_doframe(pp);

			/* State Reset. */
			pp->rcount = 0;
			pp->rhead = pp->rbuff;
			pp->escape = 0;
			pp->toss = 0;
			break;

		default:			/* regular character */
			if (! in_rmap(pp, *c)) {
				if (pp->toss == 0)
					pp_enqueue(pp, *c ^ pp->escape);
				pp->escape = 0;
			}
			break;
	}
  }
}


/*
 * Called by lower levels of TTY driver when data becomes
 * available.  All incoming data comes through this function.
 */
void
ppp_handler(struct tty_struct *tty)
{
  unsigned char buff[128];
  struct ppp *pp = pp_find(tty);
  int n, error;

  if (pp == NULL || pp->magic != PPP_MAGIC) {
	printk("PPP: handler called but couldn't find PPP struct.\n");
	return;
  }

  CHECK_PPP_VOID();

#ifdef PPP_DEBUG
  if (ppp_debug >= 5)
	ppp_dump_inqueue(pp->tty);
#endif

  do {
	n = tty_read_raw_data(tty, buff, 128);
	if (n == 0) break;	/* nothing there */

#ifdef PPP_DEBUG
	if (ppp_debug >= 5)
		ppp_print_buffer("receive buffer",
			buff, (n > 0) ? n : -n, KERNEL_DS);
#endif

	if (n < 0) {
		/*
		 * Last character is error flag.
		 * Process the previous characters, then set toss flag.
		 */
		n = (-n) - 1;
		error = buff[n];
	} else error = 0;
	pp->stats.rbytes += n;
	ppp_unesc(pp, buff, n);
	if (error) pp->toss = error;
  } while(1);
}


void
ppp_kick_tty(struct ppp *pp)
{
  register int count = pp->xhead - pp->xbuff;
  register int answer;

  pp->stats.sbytes += count;
  answer = tty_write_data(pp->tty, pp->xbuff, count,
			  ppp_output_done, (void *) pp);

  if (answer == 0) {
	ppp_output_done(pp);		/* should not happen */
  } else {
	if (answer < 0) {
		pp->stats.serrors++;
		ppp_output_done(pp);	/* unlock the transmitter */
	}
  }
}
#endif	/* NEW_TTY_DRIVERS */


/*
 * Read a PPP frame from the us_rbuff circular
 * buffer, waiting if necessary.
 */
int
ppp_read(struct tty_struct *tty, struct file *file,
	 unsigned char *buf, unsigned int nr)
{
  struct ppp *pp = pp_find(tty);
  unsigned char c;
  int len, i;

  if (pp == NULL || pp->magic != PPP_MAGIC) {
	DPRINTF(("PPP: read: cannnot find PPP channel\n"));
	return(-EIO);
  }

  CHECK_PPP(-ENXIO);

  DPRINTF(("PPP: read: called %lX num %u\n", (unsigned long) buf, nr));
  do {
	/* Try to acquire read lock. */
	if (set_bit(0, &pp->us_rbuff_lock) == 0) {
		/* Got lock. */
		if (pp->us_rbuff_head == pp->us_rbuff_tail) {
			/* No data. */
			DPRINTF(("PPP: read: no data\n"));
			clear_bit(0, &pp->us_rbuff_lock);
			if (pp->inp_sig) {
				DPRINTF(("PPP: read: EWOULDBLOCK\n"));
				return(-EWOULDBLOCK);
			} else goto wait;
		}

		/* Reset the time of the last read operation. */
		pp->ddinfo.nip_rjiffies = jiffies;

		/* Fetch packet length. */
		GETC(c); len = (c << 8);
		GETC(c); len |= (c & 0xff);
		DPRINTF(("PPP: read: len=%u\n", len));

		if ((len + 2) > nr) {
			/*
			 * Frame too big; can't copy it, but do
			 * update us_rbuff_head.
			 */
			DPRINTF(("PPP: read of %u bytes too small for %d frame\n",
							nr, len+2));
			pp->us_rbuff_head += len;
			if (pp->us_rbuff_head > pp->us_rbuff_end)
				pp->us_rbuff_head +=
					- (pp->us_rbuff_end - pp->us_rbuff);
			clear_bit(0, &pp->us_rbuff_lock);
			wake_up_interruptible(&pp->read_wait);
			pp->stats.rgiants++;
			return(-EOVERFLOW);		/* ZZZ; HACK! */
		} else {
			/*
			 * Have the space: copy the packet, faking the
			 * first two bytes.
			 */
			put_fs_byte(PPP_ADDRESS, buf++);
			put_fs_byte(PPP_CONTROL, buf++);
			i = len;
			while (i-- > 0) {
				GETC(c);
				put_fs_byte(c, buf++);
			}
		}

		clear_bit(0, &pp->us_rbuff_lock);
		DPRINTF(("PPP: read: passing %d bytes up\n", len + 2));
		pp->stats.rothers++;
		return(len + 2);
	}

wait:
	/* We need to wait. */
	current->timeout = 0;
	DPRINTF(("PPP: read: sleeping\n"));
	interruptible_sleep_on(&pp->read_wait);
	if (current->signal & ~current->blocked) return(-EINTR);
  } while(1);

  /*NOTREACHED*/
  return(0);
}


/*
 * Write a frame with NR chars from BUF to TTY.
 * We have to put the FCS field on ourselves.
 */
int
ppp_write(struct tty_struct *tty, struct file *file,
	  unsigned char *buf, unsigned int nr)
{
  struct ppp *pp = pp_find(tty);
  int i;

  if (pp == NULL || pp->magic != PPP_MAGIC) {
	DPRINTF(("PPP: write: cannot find PPP unit\n"));
	return(-EIO);
  }

  CHECK_PPP(-ENXIO);
  
  /* Has someone has been ifconfigging? */
  if (pp->mtu != pp->dev->if_mtu)
	pp_changedmtu(pp, pp->dev->if_mtu, pp->mru);

  if (nr > pp->mtu) {
	DPRINTF(("PPP: write: truncating user packet from %u to MTU %d\n",
						nr, pp->mtu));
	nr = pp->mtu;
  }

#ifdef PPP_DEBUG
  if (ppp_debug >= 3)
	ppp_print_buffer ("write frame", buf, nr, USER_DS);
#endif

  /* lock this PPP unit so we will be the only writer; sleep if necessary. */
  while((pp->sending == 1) || !pp_lock(pp)) {
	current->timeout = 0;
	DPRINTF(("PPP: write: sleeping\n"));
	interruptible_sleep_on(&pp->write_wait);
	if (current->signal & ~current->blocked) return(-EINTR);
  }

  /* OK, locked.  Stuff the given bytes into the buffer. */
  DPRINTF(("PPP: write: acquired write lock\n"));
  pp->xhead = pp->xbuff;

#ifdef OPTIMIZE_FLAG_TIME
  if (jiffies - pp->last_xmit > OPTIMIZE_FLAG_TIME) *pp->xhead++ = PPP_FLAG;
  pp->last_xmit = jiffies;
#else      
  *pp->xhead++ = PPP_FLAG;
#endif

  pp->fcs = PPP_FCS_INIT;
  i = nr;
  while (i-- > 0)
	pp_stuff_char(pp, get_fs_byte(buf++));
  ppp_add_fcs(pp);		/* concatenate FCS at end */
  *pp->xhead++ = PPP_FLAG;

  /* Reset the time of the last write operation. */
  pp->ddinfo.nip_sjiffies = jiffies;

  DPRINTF(("PPP: write: writing %d chars\n", pp->xhead - pp->xbuff));
#ifdef PPP_DEBUG
  if (ppp_debug >= 6)
	ppp_print_buffer("xmit buffer", pp->xbuff,
			pp->xhead - pp->xbuff, KERNEL_DS);
#endif

  /* Packet is ready-to-go. */
  pp->stats.sothers++;
  ppp_kick_tty(pp);

  return((int) nr);
}
 

int
ppp_select(struct tty_struct *tty, struct inode * inode,
	   struct file * filp, int sel_type, select_table * wait)
{
  struct ppp *pp = pp_find(tty);
  
  if (pp == NULL || pp->magic != PPP_MAGIC) {
	DPRINTF(("PPP: select: can't find PPP block from tty!\n"));
	return(-EBADF);
  }
  
  /* If the PPP protocol is no longer active, return false. */
  CHECK_PPP(0);
  
  /* Process the request based upon the type desired */
  switch (sel_type) {
	case SEL_IN:
		if (set_bit(0, &pp->us_rbuff_lock) == 0) {
			/* Test for the presence of data in the queue */
			if (pp->us_rbuff_head != pp->us_rbuff_tail) {
				clear_bit(0, &pp->us_rbuff_lock);
				return(1);
			}
			clear_bit(0, &pp->us_rbuff_lock);
		}
		/*FALLTHROUGH*/

	case SEL_EX:
		/* Is there a pending error condition? */
		if (tty->packet && tty->link->ctrl_status) return(1);
    
		/* closed? */
		if (tty->flags & (1 << TTY_SLAVE_CLOSED)) return(1);
    
		/* If the TTY is disconnected, then this is an exception too. */
		if (tty_hung_up_p(filp)) return(1);

		select_wait(&pp->read_wait, wait);
		break;
    
	case SEL_OUT:
		if (pp_lock(pp)) {
			if (pp->sending == 0) {
				pp_unlock(pp);
				return(1);
			}
			pp_unlock(pp);
		}
		select_wait(&pp->write_wait, wait);
		break;
  }
  return(0);
}
